from typing import Final

class Bars:
    """
Represents a collection of OHLCV (Open, High, Low, Close, Volume) data
for traditional and Heikin-Ashi bars.

The `Bars` struct contains two vectors: one for standard OHLCV data and
another for Heikin-Ashi bar data. This struct provides an organized way
to manage and manipulate historical price data, facilitating
various analytical and trading strategies.

Attributes:
    ohlcvs (Vec<OHLCV>): A vector containing the standard OHLCV data.
    ha_bars (Vec<OHLCV>): A vector containing Heikin-Ashi OHLCV data.
    """
    @property
    def ha_bars(self): ...
    @property
    def ohlcvs(self): ...
    @property
    def range_bars(self): ...
    @property
    def trades(self): ...
    def write_trades_to_parquet(self, file_path): ...
    def write_ohlcv_to_parquet(self, file_path): ...
    def build_range_bars(self, threshold, is_percentage): ...
    def add_trade(self, trade): ...
    def add_ohlcv(self, ohlcv): ...
    @staticmethod
    def __new__(*args, **kwargs) -> Bars: ...
    def __lt__(self, value): ...
    def __init__(self, *args, **kwargs): ...
    def __hash__(self): ...
    def __eq__(self, value): ...

class OHLCV:
    @property
    def close(self): ...
    @property
    def high(self): ...
    @property
    def low(self): ...
    @property
    def open(self): ...
    @property
    def symbol(self): ...
    @property
    def time(self): ...
    @property
    def timestamp_ms(self): ...
    @property
    def volume(self): ...
    @staticmethod
    def __new__(*args, **kwargs) -> OHLCV: ...
    def __lt__(self, value): ...
    def __init__(self, *args, **kwargs): ...
    def __hash__(self): ...
    def __eq__(self, value): ...

class Side:
    Long: Final[Side]
    Short: Final[Side]
    @staticmethod
    def __new__(*args, **kwargs) -> Side: ...
    def __lt__(self, value): ...
    def __int__(self): ...
    def __init__(self, *args, **kwargs): ...
    def __eq__(self, value): ...

class Trade:
    @property
    def price(self): ...
    @property
    def side(self): ...
    @property
    def size(self): ...
    @property
    def symbol(self): ...
    @property
    def time(self): ...
    @property
    def timestamp_ms(self): ...
    @staticmethod
    def __new__(*args, **kwargs) -> Trade: ...
    def __lt__(self, value): ...
    def __init__(self, *args, **kwargs): ...
    def __hash__(self): ...
    def __eq__(self, value): ...




import ccxt
import ccxt.pro
import msgspec

from tradingtoolbox.exchanges.exchange_trades import ExchangeTrades
from tradingtoolbox.utils.cache import FileFormat
from .contract import Contract
from tradingtoolbox.rs import OHLCV, Bars
from .exchange_config import ExchangeConfig
from tradingtoolbox.utils import Cache, print

from .exchange_base import ExchangeBase
from .exchange_ohlcv import ExchangeOHLCV


class Exchange(ExchangeBase, ExchangeOHLCV, ExchangeTrades):
    @staticmethod
    def available_exchanges():
        return ccxt.pro.exchanges

    @classmethod
    def create(cls, **kwargs):
        exchange_name = kwargs["exchange_name"]
        if exchange_name not in ccxt.pro.exchanges:
            raise ValueError(f"Exchange {exchange_name} not available")

        config = msgspec.json.decode(msgspec.json.encode(kwargs["config"]))
        kwargs["exchange"] = getattr(ccxt.pro, exchange_name)(config)
        self = cls(**kwargs)

        return self

    async def load_contracts(self, reload=True) -> dict:
        cache = Cache(
            cache_path=f"./cache/{self.exchange_name}_markets.json",
            method=self.exchange.load_markets,
        )
        self.contracts = {}
        data = await cache.get_async(reload=reload)

        for key in data:
            curr = Contract(**data[key])
            self.contracts[key] = curr

        return self.contracts

    async def close(self):
        await self.exchange.close()

    def find_contracts(self, symbol_name: str, contract_type: str) -> list[Contract]:
        matches = []
        for key in self.contracts:
            if symbol_name in self.contracts[key].symbol:
                contract = self.contracts[key]
                if contract.type == contract_type:
                    matches.append(contract)
        return matches



from types import ModuleType
import inspect
import os
import sys
import argparse
import importlib


class Scanner:
    scanned = set()
    classes = {}

    BLACKLIST = [
        "has_attribute",
        "__subclasshook__",
        "__class__",
        "__dir__",
        "__doc__",
        "__format__",
        "__getattribute__",
        # "__init__",
        # "__new__",
        "__init_subclass__",
        "__sizeof__",
        "__setattr__",
        "__reduce__",
        "__reduce_ex__",
        "__module__",
        "__getstate__",
        "__delattr__",
        "__ge__",
        "__gt__",
        "__le__",
        "__ne__",
        "__repr__",
        "__str__",
    ]
    WHITELIST = ["__new__", "__init__"]

    def scan_module(self, root_namespace: str, module: ModuleType):
        for element_name in dir(module):
            if element_name.startswith("__"):
                continue

            if root_namespace and root_namespace in module.__name__:
                continue

            if element_name in self.scanned:
                continue

            self.scanned.add(element_name)

            module_root = (
                f"{root_namespace + ('.' if root_namespace else '')}{module.__name__}"
            )
            class_obj = getattr(module, element_name, None)

            if isinstance(class_obj, ModuleType):
                self.scan_module(module_root, class_obj)
                continue

            class_members_all = inspect.getmembers(class_obj)
            dunder_members = [
                m
                for m in class_members_all
                if m[0].startswith("__") and m[0] not in Scanner.BLACKLIST
            ]
            class_members = [
                m
                for m in class_members_all
                if not m[0].startswith("__") and m[0] not in Scanner.BLACKLIST
            ]

            class_methods = []
            class_constants = []
            class_attributes = []
            for m in class_members:
                print(m)
                if callable(m[1]):
                    class_methods.append(m)
                elif inspect.isdatadescriptor(m[1]):
                    class_attributes.append(m)
                else:
                    class_constants.append(m)
            for m in dunder_members:
                class_methods.append(m)

            class_definition = (
                class_obj,
                class_constants,
                class_attributes,
                class_methods,
            )

            self.classes.setdefault(module_root, list())
            self.classes[module_root].append(class_definition)

    def write_pyis(self, file_location):
        for module, classes in reversed(self.classes.items()):
            print(f"Writing pyi for {module}...")
            path = module.replace(".", "/")
            path = os.path.basename(path)
            path = f"{file_location}/{path}.pyi"
            os.makedirs(os.path.dirname(f"{file_location}"), exist_ok=True)
            with open(path, "w") as file:
                file.write("from typing import Final\n\n")
                for (
                    class_obj,
                    class_constants,
                    class_attributes,
                    class_methods,
                ) in classes:
                    print(class_obj.__name__)
                    print(class_methods)
                    print(class_obj.__new__.__text_signature__)
                    print("-----------------------")
                    # init_method = getattr(class_obj, "__new__", None)
                    # try:
                    #     # init_signature = inspect.text(init_method)
                    #     init_signature = init_method.__text_signature__
                    #     print(init_signature)
                    # except ValueError:
                    #     pass

                    # init_method = getattr(class_obj, "__init__", None)
                    # try:
                    #     # init_signature = inspect.text(init_method)
                    #     init_signature = init_method.__text_signature__
                    #     print(init_signature)
                    # except ValueError:
                    #     pass

                    doc = inspect.getdoc(class_obj)

                    file.write(f"class {class_obj.__name__}:\n")
                    if doc:
                        file.write('    """\n')
                        file.write(doc)
                        file.write("\n")
                        file.write('    """\n')

                    for const in class_constants:
                        file.write(
                            f"    {const[0]}: Final[{const[1].__class__.__name__}]\n"
                        )

                    for attr in class_attributes:
                        doc = inspect.getdoc(attr[1])
                        file.write("    @property\n")
                        # file.write(f"    def {attr[0]}(self) -> {attr_type}: ...\n")
                        file.write(f"    def {attr[0]}(self): ...\n")
                        if doc:
                            file.write('    """\n')
                            file.write(doc)
                            file.write('    """\n')

                    # always put new first if present
                    class_methods.sort(
                        reverse=True,
                        key=lambda m: "aaaaa" if m[0] == "__new__" else m[0],
                    )

                    for method in class_methods:
                        if callable(method[1]):
                            try:
                                sig = inspect.signature(method[1])
                                if method[0] == "__new__":
                                    print(sig)

                                if sig.return_annotation is not inspect._empty:
                                    print(f"RETURN  = {sig.return_annotation}")
                                if "self" not in sig.parameters:
                                    file.write("    @staticmethod\n")

                                file.write(
                                    f"    def {method[0]}({', '.join([str(p) for p in sig.parameters.values()])})"
                                )
                            except ValueError:
                                file.write(
                                    f"    def {method[0]}{method[1].__text_signature__}"
                                )
                            if method[0] == "__new__":
                                file.write(f" -> {class_obj.__name__}")
                            file.write(": ...\n")
                            if not method[0].startswith("__"):
                                doc = inspect.getdoc(method[1])
                                if doc:
                                    file.write('    """\n')
                                    file.write(doc)
                                    file.write("\n")
                                    file.write('    """\n')

                    file.write("\n")


def main():
    parser = argparse.ArgumentParser(
        description="Scan a module and generate .pyi files"
    )
    parser.add_argument(
        "module", help="The name of the module to scan (e.g., zenbt.rs)"
    )
    parser.add_argument("module_dir", help="The directory where the module is located")
    args = parser.parse_args()

    # Add module_dir to sys.path
    sys.path.append(args.module_dir)
    print(args)

    # Dynamically import the module
    try:
        module = importlib.import_module(args.module)
        print(module)
    except ModuleNotFoundError:
        print(f"Module {args.module} not found in {args.module_dir}")
        return

    module_file = module.__file__
    if module_file is None:
        print(
            f"Module {args.module} has no __file__ attribute (might be a built-in or extension module)."
        )
        return

    # Instantiate the scanner and scan the module
    scanner = Scanner()
    scanner.scan_module("", sys.modules[args.module])

    # Write the .pyi file in the same directory as the module
    # module_dir = os.path.dirname(module_file)
    scanner.write_pyis(args.module_dir)


if __name__ == "__main__":
    main()



import ccxt
import ccxt.pro
import msgspec

from tradingtoolbox.utils.cache import FileFormat
from .contract import Contract
from tradingtoolbox.rs import OHLCV, Bars
from .exchange_config import ExchangeConfig
from tradingtoolbox.utils import Cache, print


class ExchangeOHLCV:
    config: ExchangeConfig
    exchange_name: str
    exchange: ccxt.pro.Exchange
    contracts: dict[str, Contract] = {}
    file_name: str = ""

    async def fetch_historical_ohlcv(
        self, contract: Contract, timeframe="1d", pages=3, reload=False
    ) -> Bars:
        """
        Fetches historical OHLCV (Open, High, Low, Close, Volume) data for a specified contract.

        This asynchronous method retrieves the specified number of historical candle data pages
        for a given contract and timeframe. It can be used for various candle manipulation operations.

        Attributes:
            contract (Contract): The contract for which historical data is being fetched.
            timeframe (str): The time interval for the OHLCV data (default is "1d").
            pages (int): The number of pages of data to fetch (default is 3).
            reload (bool): Flag indicating whether to reload data if it exists (default is False).

        Returns:
            Bars: An instance of Bars containing the historical OHLCV data.

        Examples:
            >>> bars = await fetch_historical_ohlcv(contract, "1h", pages=5)
            >>> print(bars)
        """
        file_name = (
            f"./data/{self.exchange_name}_{contract.id}_{timeframe}_ohlcv.parquet"
        )
        cache = Cache(
            file_format=FileFormat.PARQUET,
            dump_format=FileFormat.NONE,
            cache_path=file_name,
            method=self._fetch_historical_ohlcv,
        )
        data = await cache.get_async(
            reload=reload, contract=contract, timeframe=timeframe, pages=pages
        )

        bars = Bars()
        for i in range(len(data)):
            candle = data.iloc[i]
            bar = OHLCV(
                contract.symbol,
                candle.timestamp_ms,
                candle.open,
                candle.high,
                candle.low,
                candle.close,
                candle.volume,
            )
            bars.add_ohlcv(bar)

        return bars

    async def _fetch_historical_ohlcv(
        self, contract: Contract, timeframe="1d", pages=3
    ) -> Bars:
        file_name = (
            f"./data/{self.exchange_name}_{contract.id}_{timeframe}_ohlcv.parquet"
        )
        print("Fetching historical")

        candles = await self.exchange.fetch_ohlcv(
            contract.symbol,
            timeframe,
            params={"paginate": True, "paginationCalls": pages},
        )

        if len(candles) > 0:
            bars = Bars()
            for candle in candles:
                bar = OHLCV(contract.symbol, *candle)
                # print(bar)
                bars.add_ohlcv(bar)
            bars.write_ohlcv_to_parquet(file_name)



# from .binance import BinanceKlines, Timeframes
# from .okx import OKXKlines
# __all__ = ["BinanceKlines", "Timeframes", "OKXKlines", "Exchange", "ExchangeConfig"]

from .exchange import Exchange
from .exchange_config import ExchangeConfig


__all__ = ["Exchange", "ExchangeConfig"]



import msgspec
from typing import Optional


class ExchangeConfig(msgspec.Struct):
    enableRateLimit: bool
    apiKey: Optional[str] = ""
    secret: Optional[str] = ""
    password: Optional[str] = ""
    uid: Optional[str] = None
    newUpdates: Optional[bool] = None
    options: dict = {}

    @classmethod
    def create(cls, config_file_path):
        with open(config_file_path, "rb") as f:
            config = msgspec.json.decode(f.read())
        return cls(**config)



"""
A bunch of util functions
"""

from .logger import logger, print, Logger
from .cache import Cache
from .time_manip import TimeManip
from .resample import resample

__all__ = ["logger", "Logger", "print", "Cache"]



import ccxt
import ccxt.pro
import msgspec
from datetime import datetime

from tradingtoolbox.utils.cache import FileFormat
from .contract import Contract
from tradingtoolbox.rs import Bars, Trade, Side
from .exchange_config import ExchangeConfig
from tradingtoolbox.utils import Cache, print


class ExchangeTrades:
    config: ExchangeConfig
    exchange_name: str
    exchange: ccxt.pro.Exchange
    contracts: dict[str, Contract] = {}
    file_name: str = ""

    async def fetch_historical_trades(
        self, contract: Contract, since, reload=False
    ) -> Bars:
        self.file_name = f"./data/{self.exchange_name}_{contract.id}_trades.parquet"
        cache = Cache(
            file_format=FileFormat.PARQUET,
            dump_format=FileFormat.NONE,
            cache_path=self.file_name,
            method=self._fetch_historical_trades,
        )
        # cache.clean_cache()
        data = await cache.get_async(reload=reload, contract=contract, since=since)

        bars = Bars()
        for i in range(len(data)):
            trade = data.iloc[i]
            side = Side.Long if trade.side == "buy" else Side.Short
            bar = Trade(
                contract.symbol,
                trade.timestamp_ms,
                trade.price,
                trade.size,
                side,
            )
            bars.add_trade(bar)

        return bars

    async def _fetch_historical_trades(self, contract: Contract, since) -> Bars:
        trades = await self.exchange.fetch_trades(
            contract.symbol,
            since=since,
            params={"paginate": True, "paginationDirection": "backward"},
        )  # dynamic/time-based pagination starting from 1664812416000
        print(len(trades))
        # # since = 1664812416000
        # d = datetime.fromtimestamp(since)
        # trades = []
        # while since < self.exchange.milliseconds():
        #     limit = 10000  # change for your limit
        #     _trades = await self.exchange.fetch_trades(contract.symbol, since, limit)
        #     if len(_trades) > 1:
        #         since = int(_trades[-1]["timestamp"])
        #         d = datetime.fromtimestamp(since / 1000)
        #         trades += _trades
        #     else:
        #         break
        # print(len(trades))

        if len(trades) > 0:
            bars = Bars()
            for trade in trades:
                side = Side.Long if trade["side"] == "buy" else Side.Short
                trade = Trade(
                    contract.symbol,
                    timestamp_ms=trade["timestamp"],
                    price=trade["price"],
                    size=trade["amount"],
                    side=side,
                )
                bars.add_trade(trade)
            bars.write_trades_to_parquet(self.file_name)



from tradingtoolbox.utils import print
import pytest

from tradingtoolbox.clickhouse import ClickhouseSync, ClickhouseAsync

from tradingtoolbox.utils.time_manip import time_manip

import pandas as pd
import asyncio
import uvloop
import msgspec
import ccxt
import ccxt.pro
from tradingtoolbox.utils import print
from tradingtoolbox.exchanges import Exchange, ExchangeConfig
from tradingtoolbox.utils import Cache
# from tradingtoolbox.exchanges.binance import BinanceKlines, Timeframes


def test_clickhouse():
    print("Hello")

    # logger.warning("Warning")
    # logger.error("Error message")

    a = 4
    assert a == 4


#     bt = BinanceKlines()
#     df = bt.get_futures_klines(
#         Timeframes.TF_1HOUR, asset="BTCUSDT", ago="1 day ago UTC"
#     )
#     print(df)

#     klines = OKXKlines()
#     df = klines.load_klines("PEPE-USDT-SWAP", "1m", days_ago=30)
#     print(df)

#     ch = ClickhouseSync.create()
#     a = ch.client.command("SELECT 1")
#     print(a)
#     # self.clickhouse = await Clickhouse.create()
#     dic = {"test": "123"}
#     print("Hello wotld", 123, dic)

#     async def main():
#         ch = await ClickhouseAsync.create()
#         a = await ch.async_client.command("SELECT 1")
#         print(a)

#     asyncio.run(main())

#     days = time_manip.days_ago(3)
#     print(days)



import os
import json
from msgspec.json import decode
import msgspec
import asyncio
import pandas as pd
from datetime import date
from typing import Callable, Optional
from enum import Enum


class FileFormat(Enum):
    JSON = "json"
    PARQUET = "parquet"
    NONE = "none"


class Cache(msgspec.Struct):
    """
    Cache class for handling file-based caching with support for asynchronous data loading.

    Attributes:
        cache_path (str): The path where the cache file is stored.
        method (Optional[Callable]): A callable, typically an async method, to fetch data when the cache doesn't exist or needs to be reloaded.
        date_as_suffix (bool): If True, appends today's date to the cache file name as a suffix.

    Methods:
        get_async(reload=True): Asynchronously retrieves data from the cache. If the cache file doesn't exist, it will fetch data using the provided method and store it in the cache.
        clean_cache(): Deletes the cache file from disk.

    Usage:
    ```py
    async def my_function():
        await asyncio.sleep(5)
        return []

    cache = Cache(
        cache_path=f"./cache/markets_{ex.exchange_name}.json", method=ex.get_contracts
    )
    data = await cache.get_async()

    # Use this if you don't have any other async.io task, otherwise the program will end before `my_function` runs
    await cache.wait_till_complete()
    ```
    """

    # TODO
    # 1. Provide custom method to load the cache instead of assuming it's json
    # 2. allow to pass parameters to the method

    cache_path: str
    """Where you want the file to be stored"""
    method: Optional[Callable] = None
    """A lambda method to run in case the file doesn't exist"""
    date_as_suffix: bool = False
    """If true, will append today's date to the cache file."""
    file_format: FileFormat = FileFormat.JSON
    """Format of the file to read."""
    dump_format: FileFormat = FileFormat.JSON
    """Format of the file to dump to. It might be None if your method already dumps the file."""
    _task: Optional[asyncio.Task] = None

    def __post_init__(self):
        # Custom initialization logic after the msgspec-generated init
        if self.date_as_suffix:
            extension = self.cache_path.split(".")[-1]
            name_without_extension = self.cache_path.replace(f".{extension}", "")
            self.cache_path = f"{name_without_extension}_{date.today().strftime('%Y%m%d')}.{extension}"
        self._ensure_path_exists()

    # ====================== Private Methods ======================= #
    def _read_file(self) -> Optional[dict | pd.DataFrame]:
        if os.path.exists(self.cache_path):
            if self.file_format == FileFormat.JSON:
                with open(self.cache_path, "r") as f:
                    return decode(f.read())
            elif self.file_format == FileFormat.PARQUET:
                return pd.read_parquet(self.cache_path)
        return None

    def _ensure_path_exists(self):
        directory_path = os.path.dirname(self.cache_path)
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)

    async def _reload_async(self, **method_kwargs):
        if self.method:
            data = await self.method(**method_kwargs)

            if self.dump_format == FileFormat.JSON:
                with open(self.cache_path, "w") as f:
                    json.dump(data, f, indent=2)
            elif self.dump_format == FileFormat.PARQUET:
                data.to_parquet(self.cache_path, index=False)
            elif self.dump_format == FileFormat.PARQUET:
                pass

    async def wait_till_complete(self):
        """
        Waits for the task to complete.
        Use this if you don't have any other async.io task to avoid the program ending before the method is ran
        """
        if self._task:
            while True:
                if self._task.done():
                    break
                await asyncio.sleep(1)  # Check every second

    @staticmethod
    async def wait_till_all_tasks_complete():
        """
        Waits for all tasks to complete.
        Use this if you don't have any other async.io task to avoid the program ending before the method is ran
        """
        while True:
            pending_tasks = []
            for task in asyncio.all_tasks():
                if not task.done():
                    pending_tasks.append(task)
            if len(pending_tasks) == 1:
                break
            await asyncio.sleep(1)

    # ====================== Public Methods ======================= #
    async def get_async(self, reload=True, **method_kwargs):
        """
        Gets the data from the cache. If it doesn't exist, it will reload it.

        Args:
            reload (bool, optional): Whether to reload the data. Defaults to True.
            **method_kwargs: Keyword arguments to pass to the method.

        Returns:
            dict: The data from the cache.
        """
        data = self._read_file()
        if data is not None:
            if reload:
                self._task = asyncio.create_task(self._reload_async(**method_kwargs))
        else:
            await self._reload_async(**method_kwargs)
            data = self._read_file()
        return data  # type: ignore

    def clean_cache(self):
        """
        Removes the cache file.
        """
        if os.path.exists(self.cache_path):
            os.remove(self.cache_path)



from datetime import datetime
from enum import Enum
from typing import Optional, Dict, Any
from decimal import Decimal as D
from typing import Union
import math
import msgspec


def normalize_fraction(decim: D) -> D:
    normalized = decim.normalize()
    sign, digit, exponent = normalized.as_tuple()
    return normalized if exponent <= 0 else normalized.quantize(1)


def decimal_from_value(value: Union[str, int, float]) -> D:
    return normalize_fraction(D(str(value)))


class Precision(msgspec.Struct):
    amount: float
    price: float
    cost: Optional[float] = None
    base: Optional[str] = None
    quote: Optional[str] = None


class Limits(msgspec.Struct):
    leverage: Dict[str, Optional[float]]
    amount: Dict[str, Optional[float]]
    price: Dict[str, Optional[float]]
    cost: Dict[str, Optional[float]]


class LotSizeFilter(msgspec.Struct):
    maxOrderQty: Optional[float] = None
    minOrderQty: float = 0
    maxOrderAmt: Optional[float] = None
    minOrderAmt: Optional[float] = None
    maxMktOrderQty: Optional[float] = None
    basePrecision: Optional[float] = None
    quotePrecision: Optional[float] = None
    qtyStep: float = 0
    postOnlyMaxOrderQty: Optional[float] = None
    minNotionalValue: Optional[float] = None

    def __post_init__(self, **kwargs):
        for key in kwargs:
            try:
                setattr(self, key, float(kwargs[key]))
            except ValueError:
                setattr(self, key, 0)

        if "qtyStep" not in kwargs and self.basePrecision:
            self.qtyStep = self.basePrecision


class PriceFilter(msgspec.Struct):
    minPrice: float = 0
    maxPrice: float = 0
    tickSize: float = 0

    def __post_init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, float(kwargs[key]))


class BybitInfo(msgspec.Struct):
    symbol: str
    baseCoin: str
    quoteCoin: str
    status: str
    priceFilter: PriceFilter
    lotSizeFilter: LotSizeFilter
    innovation: Optional[str] = None
    settleCoin: Optional[str] = None
    optionsType: Optional[str] = None
    launchTime: Optional[int] = None
    deliveryTime: Optional[int] = None
    deliveryFeeRate: Optional[str] = None


class OkxInfo(msgspec.Struct):
    symbol: str
    baseCoin: str
    quoteCoin: str
    status: str
    priceFilter: PriceFilter
    lotSizeFilter: LotSizeFilter
    innovation: Optional[str] = None
    settleCoin: Optional[str] = None
    optionsType: Optional[str] = None
    launchTime: Optional[int] = None
    deliveryTime: Optional[int] = None
    deliveryFeeRate: Optional[str] = None


class FeeSide(str, Enum):
    get = "get"


class Contract(msgspec.Struct, kw_only=True):
    id: str
    symbol: str
    base: str
    quote: str
    lowercaseId: Optional[str] = None
    settle: Optional[str] = None
    baseId: str
    quoteId: str
    settleId: Optional[str] = None
    type: str
    spot: bool
    margin: Optional[bool] = None
    swap: bool
    future: bool
    option: bool
    index: Optional[bool] = None
    active: bool
    contract: bool
    linear: Optional[bool] = None
    inverse: Optional[bool] = None
    subType: Optional[str] = None
    taker: float
    maker: float
    contractSize: Optional[float] = None
    expiry: Optional[int] = None
    expiryDatetime: Optional[datetime] = None
    strike: Optional[float] = None
    optionType: Optional[str] = None
    precision: Precision
    limits: Limits
    created: Optional[int] = None
    info: Any
    tierBased: Optional[bool] = None
    percentage: Optional[bool] = None
    feeSide: Optional[FeeSide] = None
    marginModes: dict

    def round_value(self, _value: float, _step: float, how: str = "down") -> float:
        value = decimal_from_value(_value)
        step = decimal_from_value(_step)
        if how == "down":
            return float(value // step * step)
        elif how == "up":
            return float(math.ceil(value / step) * step)
        else:
            return float(round(value / step) * step)

    def round_size(self, size: float) -> float:
        if self.limits.amount["min"] is None:
            return size

        if size < self.limits.amount["min"]:
            size = self.limits.amount["min"]
            print(
                f"Size is smaller than min size. Reverting to the min allowed which is {size}"
            )
        return self.round_value(size, self.limits.amount["min"])

    def round_price(self, price: float) -> float:
        return self.round_value(price, self.precision.price)



import ccxt
import ccxt.pro
import msgspec

from tradingtoolbox.utils.cache import FileFormat
from .contract import Contract
from tradingtoolbox.rs import OHLCV, Bars
from .exchange_config import ExchangeConfig
from tradingtoolbox.utils import Cache, print


class ExchangeBase(msgspec.Struct):
    config: ExchangeConfig
    exchange_name: str
    exchange: ccxt.pro.Exchange
    contracts: dict[str, Contract] = {}
    file_name: str = ""



import pytest

from tradingtoolbox.clickhouse import ClickhouseSync, ClickhouseAsync

from tradingtoolbox.utils.time_manip import time_manip

import pandas as pd
import asyncio
import uvloop
import msgspec
import ccxt
import ccxt.pro
from tradingtoolbox.utils import print
from tradingtoolbox.exchanges import Exchange, ExchangeConfig
from tradingtoolbox.utils import Cache
# from tradingtoolbox.exchanges.binance import BinanceKlines, Timeframes


async def init() -> Exchange:
    print("")
    # config = ExchangeConfig.create(config_file_path="./config/bybit_config.json")
    # ex = Exchange.create(exchange_name="bybit", config=config)

    config = ExchangeConfig.create(config_file_path="./config/okx_config.json")
    ex = Exchange.create(exchange_name="okx", config=config)

    await ex.load_contracts(reload=False)
    assert len(ex.contracts) == 2392
    return ex


@pytest.mark.asyncio
async def test_ohlc():
    ex = await init()

    contracts = ex.find_contracts(symbol_name="PEPE/USDT:USDT", contract_type="swap")
    for contract in contracts:
        candles = await ex.fetch_historical_ohlcv(
            contract, timeframe="1m", pages=1, reload=False
        )
        print(candles)
        ohlcv = candles.ohlcvs
        print(ohlcv[0].open)
        print(type(candles))

    await Cache.wait_till_all_tasks_complete()
    await ex.close()


@pytest.mark.asyncio
async def test_trades():
    ex = await init()
    print("Testing trades")
    contracts = ex.find_contracts(symbol_name="PEPE/USDT:USDT", contract_type="swap")
    for contract in contracts:
        bars = await ex.fetch_historical_trades(
            contract,
            since=time_manip.minutes_ago(30, to_timestamp=True),
            reload=True,
        )
        print(bars.trades[0].time)
        print(bars.trades[-1].time)
        print(len(bars.trades))
        bars.build_range_bars(threshold=0.26, is_percentage=True)
        print(len(bars.range_bars))

    # await asyncio.sleep(10)
    await Cache.wait_till_all_tasks_complete()
    await ex.close()



from tradingtoolbox.utils import print
from tradingtoolbox.clickhouse import ClickhouseSync, ClickhouseAsync

from tradingtoolbox.utils.time_manip import time_manip

import pandas as pd
import asyncio
import uvloop
import msgspec
import ccxt
import ccxt.pro
from tradingtoolbox.utils import print
from tradingtoolbox.exchanges import Exchange, ExchangeConfig
from tradingtoolbox.utils import Cache
# from tradingtoolbox.exchanges.binance import BinanceKlines, Timeframes


# def _dev():
#     bt = BinanceKlines()
#     df = bt.get_futures_klines(
#         Timeframes.TF_1HOUR, asset="BTCUSDT", ago="1 day ago UTC"
#     )
#     print(df)

#     klines = OKXKlines()
#     df = klines.load_klines("PEPE-USDT-SWAP", "1m", days_ago=30)
#     print(df)

#     ch = ClickhouseSync.create()
#     a = ch.client.command("SELECT 1")
#     print(a)
#     # self.clickhouse = await Clickhouse.create()
#     dic = {"test": "123"}
#     print("Hello wotld", 123, dic)

#     async def main():
#         ch = await ClickhouseAsync.create()
#         a = await ch.async_client.command("SELECT 1")
#         print(a)

#     asyncio.run(main())

#     days = time_manip.days_ago(3)
#     print(days)


async def main():
    # config = ExchangeConfig.create(config_file_path="./config/okx_config.json")
    # ex = Exchange.create(exchange_name="okx", config=config)

    config = ExchangeConfig.create(config_file_path="./config/bybit_config.json")
    ex = Exchange.create(exchange_name="bybit", config=config)

    await ex.load_contracts(reload=False)

    contracts = ex.find_contracts(symbol_name="PEPE/USDT:USDT", contract_type="swap")
    for contract in contracts:
        candles = await ex.fetch_historical_ohlcv(
            contract, timeframe="1m", pages=1, reload=False
        )
        print(candles)
        ohlcv = candles.ohlcvs
        print(ohlcv[0].open)
        print(type(candles))

    await Cache.wait_till_all_tasks_complete()
    await ex.close()


async def _dev():
    await asyncio.gather(
        asyncio.create_task(main(), name="Main"),
    )


def dev():
    print("In dev")

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    asyncio.run(_dev())


dev()
